--  Package Monitoring is use for instrumenting the package Chop
--  counting the number of accepted requests when one philosopher already eats
--  counting the number of denials when one philosopher already eats
--  a hungry philosopher may meet several refusals and only one acceptance
--  deriving a ratio of concurrency denials

-- with Global; use Global; -- already declared in the specification part
with Ada.Text_IO;  use Ada.Text_IO;
with Ada.Calendar; use Ada.Calendar;
package body Monitoring is
	
	-- Global has declared 
	-- N : constant := 5; -- number of philosophers
    -- type Id is mod N;
	-- Next_Hazard : a function providing a random number
	-- Display_Lock : providing mutual exclusion for results displaying

	--      ****** simulation lenght control 

	MaxStrokes : constant Integer :=  100; -- 100000; -- simulation lenght
	protected Counter is
		procedure Reset;
		procedure Increment(Finished : out Boolean); -- needs mutual exclusion
	private
		Number : Integer := 0;
	end Counter;
	protected body Counter is
		procedure Reset is
		begin Number := 0; end Reset;
		procedure Increment(Finished : out Boolean) is
		begin 
			Number := Number + 1; Finished := Number > MaxStrokes;
		end Increment;
	end Counter;

	procedure AnotherStroke(Finished : out Boolean) is
	begin Counter.Increment(Finished); end AnotherStroke;

	--     *****   philosophers actions simulation for diner procedure

	ThinkingDuration : constant Duration := 0.01;
	EatingDuration : constant Duration := 0.01;
	 RequestingDuration : constant Duration := 0.001; -- larger value
	-- RequestingDuration : constant Duration := 0.00001; -- medium value
	-- RequestingDuration : constant Duration := 0.0; -- zero value

	-------------------        displaying messages
	procedure Message(Who : in Id; Mess : in String) is
	begin
 		null;
		--   Display_Lock.Acquire;
		--   Put_Line("Philosopher " & Id'Image(Who) &  "  " & Mess);
		--   Display_Lock.Release;
	end Message;

	procedure Thinking(Philo : in Id) is
		ThinkingDelay : Duration := ThinkingDuration*Duration(Global.Next_Hazard);
    begin
		If Philo = 1 or Philo = 4 then ThinkingDelay := 0.0; end if;
		Message (Philo, "is thinking during " & Duration'Image(ThinkingDelay));
		delay(ThinkingDelay);
		Message (Philo, "is hungry");
	end Thinking;

	procedure Eating(Philo : in Id) is
		EatingDelay : Duration := EatingDuration*Duration(Global.Next_Hazard);    
	begin
		If Philo = 1 or Philo = 4 then EatingDelay := 20* EatingDelay; end if;
		Message (Philo, "is eating during " & Duration'Image(EatingDelay));
		delay(EatingDelay);
		Message (Philo, "has finished eating");
	end Eating;

	procedure Quitting(Philo : in Id)is
    begin
		Message (Philo, "has released its chopsticks");
	end Quitting;

	procedure Leaving(Philo : in Id) is
    begin
		Message (Philo, "is leaving the group and says GOOD BYE");
	end Leaving;

	--         *** Java specific for Java notifying behaviour measurement
	procedure Requesting_LR(Philo : in Id; Result : Integer) is
		RequestingDelay : Duration := RequestingDuration*Duration(Global.Next_Hazard);
	begin
		Message(Philo, "  has received " & Integer'Image(Result) & " sticks");
		delay(RequestingDelay);
	end Requesting_LR;
	procedure Requesting_R(Philo : in Id; Result : Integer) is
		RequestingDelay : Duration := RequestingDuration*Duration(Global.Next_Hazard);
	begin
		Message(Philo, "  has received " & Integer'Image(Result) & " sticks");
		delay(RequestingDelay);
	end Requesting_R;

	--      ****      instrumentation  for chop package

	type THE is (Thinking, Hungry, Eating);
	type SticStatus is array(Id) of THE;
	Status : SticStatus  := (others => Thinking); -- philosophers states
	type PhiloStats is array(Id) of Integer ; -- statistics counter
	IndividualSuccesses : Philostats := (others => 0); -- individual successes
	IndividualStructuralRefusals : Philostats := (others => 0); -- individual structural refusals
	IndividualCautiousRefusals : Philostats:= (others => 0); -- individual cautious refusals
	CurrentlyEating : Integer := 0; -- number of concurrently eating current philosophers
	NbPairs : Integer := 0;         -- number of times two philosophers eat jointly
	NbRequestSingletons : Integer := 0; -- number of times lonely dinner
	NbStructuralRefusals : Integer := 0; 
	-- number of times lonely dinner while other hungry blocked by a neigbourgh eating
	NbCautiousRefusals : Integer := 0;
	-- number of times lonely dinner while other hungry blocked by service denial
	-- once dining, a philosopher can eat alone for a while and with another one some times
	CurrentlyHungry : Integer := 0; -- number of concurrently hungry and queued
	Tick : Integer :=0;    -- number of queue length change
	SigmaQueued : Integer := 0; -- cumulative sum of queue length

	procedure InstrumentGetSuccessful(C : in Id) is
	begin
		if Status(C) = Hungry then 	
			CurrentlyHungry := CurrentlyHungry - 1; -- a hungry philosopher leaves the queues
			SigmaQueued :=  SigmaQueued + CurrentlyHungry; -- cumulative for mean value 
			Tick := Tick + 1; -- a significative event for queuing length
		end if;
		Status(C) := Eating; IndividualSuccesses(C) := IndividualSuccesses(C) + 1; -- counts C successes
		CurrentlyEating := CurrentlyEating + 1;
		if CurrentlyEating = 2 then 
			-- two philosophers eat jointly,i.e., maximum concurrency
			NbPairs := NbPairs + 1; 
		else
			-- requesting and eating lonely (since there is no deadlock)
			-- no interest for concurrency analysis
			NbRequestSingletons :=NbRequestSingletons + 1;
		end if;
	end InstrumentGetSuccessful;

	procedure InstrumentGetRefusal(C : in Id) is
	begin
		if Status(C) = Thinking then
			CurrentlyHungry := CurrentlyHungry + 1; -- a hungry philosopher enters the queues
			SigmaQueued :=  SigmaQueued + CurrentlyHungry; -- cumulative for mean value 
			Tick := Tick + 1; -- a significative event for queuing length
		end if;
		Status(C) := Hungry;
		-- checks refusal condition and records it
		if CurrentlyEating = 1 then 
			if (Status(C+1) /= Eating) and (Status(C-1) /= Eating) then
				-- cautious denial since no neighbour is eating
				--  counts as a concurrency denial
				NbCautiousRefusals := NbCautiousRefusals + 1;
				IndividualCautiousRefusals(C) := IndividualCautiousRefusals(C) + 1; -- counts C refusals
			else
				-- structural impossibility, cannot be counted as concurrency denial
				NbStructuralRefusals := NbStructuralRefusals + 1;
				IndividualStructuralRefusals(C) := IndividualStructuralRefusals(C) + 1; -- counts C refusals
			end if;
		end if;
	end InstrumentGetRefusal;

	procedure InstrumentRelease(C : in Id) is
	begin
		Status(C) := Thinking; CurrentlyEating := CurrentlyEating - 1;
	end InstrumentRelease;

	--        **** simulation run metrics

	StartDate, StopDate : Time;

	procedure TimeStart is begin StartDate := Clock; end TimeStart;

	procedure Print_Measures;

	procedure TimeStop is 
	begin 
		StopDate := Clock; 
		Print_Measures;
	end TimeStop;

	type TimeArray is array(Id) of Time;
	ChopStartDate : TimeArray;
	type DelayArray is array(Id) of Duration;
	WaitingTotal : DelayArray := (others => 0.0);

	procedure ChopStart(Philo : in Id) is 
	--  starting service time for an allocation
	begin ChopStartDate(Philo) := Clock; end ChopStart;

	procedure ChopEnd(Philo : in Id) is
 	-- ending  service time for an allocation, adds duration to total 
	begin 
		WaitingTotal(Philo) := WaitingTotal(Philo) + (Clock - ChopStartDate(Philo));
	end ChopEnd;


	procedure Print_Measures is
		TotalRunTime, TotalWaitingTime : Duration;	
		TotalDenials : Integer; -- total of denials
		NbMealsAttended : Integer; -- Number of meals attended by philosophers
	begin
		-- concurrency effictiveness
		Put_Line (Integer'Image(Global.N) 
				& " philosophers thinking" & Duration'Image(ThinkingDuration)
				& " and eating" & Duration'Image(EatingDuration));
		Put_Line( " except Philosophers 1 and 4 with null thinking duration,"
					& " and eating duration 20 times longer than others");
		Put_Line( " this causes almost starvation when global allocation is implemented.");
		Put_Line( " (Additinally when simulating Java monitor semantics) requesting is " 
					& Duration'Image(RequestingDuration));

		--  total of concurrency denials
		TotalDenials := NbCautiousRefusals + NbStructuralRefusals;
		NbMealsAttended := NbPairs + NbRequestSingletons;
		Put_line("                 CONCURRENCY MEASURES DISPLAY"); 
		Put_line(" NbMealsAttended =  " & Integer'Image(NbMealsAttended));
		Put_Line(" NbPairs : " & Integer'Image(NbPairs));
		Put_Line(" NbRequestSingletons : " & Integer'Image(NbRequestSingletons)); 
		Put_Line(" NbStructuralRefusals : " & Integer'Image(NbStructuralRefusals)); 
		Put_Line(" NbCautiousRefusals : " & Integer'Image(NbCautiousRefusals));
		Put_Line(" Total number of denials =  " & Integer'Image(TotalDenials));
		Put_Line(" Mean hungry waiting queue length : " & Float'Image(Float(SigmaQueued/Tick)));

		-- durations measured
		TotalRunTime := StopDate - StartDate; 
		TotalWaitingTime := 0.0;
		for Philo in Id loop 
			TotalWaitingTime := TotalWaitingTime + WaitingTotal(Philo);
		end loop;
		TotalWaitingTime := TotalWaitingTime/Duration(Global.N);
		Put_Line("      Total simulation time is   " & Duration'Image(TotalRunTime));
		Put_Line("      Mean allocation time is    " & Duration'Image(TotalWaitingTime));
		Put_Line("      Allocation ratio is        " & Duration'Image(TotalWaitingTime/TotalRunTime));
		for I in Id loop
			Put_Line("      Philosopher " & Id'Image(I) & "  remains hungry during " 
							& Duration'Image(WaitingTotal(I)));
			Put_Line("        " & Integer'Image(IndividualCautiousRefusals(I)) & " cautious refusals,  "
							& Integer'Image(IndividualStructuralRefusals(I)) & " structural refusals,  "
							& Integer'Image(IndividualSuccesses(I)) & " successes");
		end loop;
	end Print_Measures;

end Monitoring; 